"""Functions for manipulating crunch cubes."""

import six

from pycrunch import elements


def fetch_cube(dataset, dimensions, weight=None, filter=None, **measures):
    """Return a shoji.View containing a crunch:cube.

    The dataset entity is used to look up its views.cube URL.
    The dimensions must be a list of either strings, which are assumed to be
    URL's of variable Entities to be fetched and analyzed according to type,
    or objects, which are assumed to be complete variable expressions.
    The weight, if sent, should be the URL of a valid weight variable
    If applying a filter, it should be a filter expression or filter URL.

    >>> dataset = session.site.datasets.by('name')['my dataset'].entity
    >>> variables = dataset.variables.by('alias')
    >>> dimensions = [
    ... {"each": variables['CA'].entity_url},
    ...     {"variable": variables['CA'].entity_url}
    ... ]
    >>> weight = variables['weight_var'].entity_url
    >>> count = {
    ...     "function": "cube_count",
    ...     "args": []
    ... }
    >>> filter = {
    ...     "function": "!=",
    ...     "args": [
    ...         {"variable": variables['categorical_var'].entity_url},
    ...         {"value": 3},
    ...     ]
    ... }
    >>> fetch_cube(dataset, dimensions, weight=weight, filter=filter, count=count)

    """
    dims = DimensionsPreparer(dataset).prepare_dimensions(dimensions)
    cube_query = elements.JSONObject(dimensions=dims, measures=measures)

    if weight is not None:
        cube_query['weight'] = weight

    params = {"query": cube_query.json}
    if filter is not None:
        params['filter'] = elements.JSONObject(filter).json

    return dataset.session.get(
        dataset.views.cube,
        params=params
    ).payload


class DimensionsPreparer(object):

    """Implement basic preparation of requested cube dimensions.

    Cube dimensions can be requested in different forms:

    1. Crunch expressions
    2. Variable URLs (or full Subvariable URLs)
    3. Variable Names
    4. Variable Aliases

    If Crunch expression is provided, no action needs to be taken, because the
    cube api can handle this notation. In the case of URLs, names, or aliases,
    the conversion needs to be done (to crunch expression), in order for cube
    api to be able to handle it.
    """

    def __init__(self, dataset):
        self._dataset = dataset
        self._variables_by_alias = dataset.variables.by('alias')
        self._variables_by_name = dataset.variables.by('name')

    def prepare_dimensions(self, dimensions):
        """Return list of crunch expressions for each cube dimension.

        :param dimensions: Collection of cube dimensions (as Crunch
            expressions, URLs, Names, or Aliases')
        :type dimensions: list of strings or dicts
        """

        dims = []

        for dim in dimensions:
            if isinstance(dim, dict):
                # This is already a Crunch expression.
                dims.append(dim)
            elif isinstance(dim, six.string_types):
                dim = self.get_dimension_by_string(dim)
                dims.extend(self.prepare_ref(dim))
            else:
                raise TypeError(
                    "Dimensions must be URL strings "
                    "or Crunch expression objects."
                )

        return dims

    def get_dimension_by_string(self, dim_str):
        """Return variable object from pycrunch dataset.

        :param dim_str: String representing URL, Name, or Alias of a variable
        """

        if dim_str in self._dataset.variables.index:
            # When URL is provided, fetch variable from index
            return self._dataset.variables.index[dim_str]
        elif dim_str in self._variables_by_alias:
            return self._variables_by_alias[dim_str]
        elif dim_str in self._variables_by_name:
            return self._variables_by_name[dim_str]
        elif 'subvariables/' in dim_str:
            var_url = dim_str.split('subvariables/')[0]
            variable = self._dataset.variables.index[var_url]
            return variable.entity.subvariables.index[dim_str]

        raise ValueError("Can't find variable {} in dataset {}".format(
            dim_str, self._dataset.self
        ))


    @staticmethod
    def prepare_ref(dim):
        ref = {'variable': dim.entity_url}
        if dim.type == "numeric":
            ref = [{"function": "bin", "args": [ref]}]
        elif dim.type == "datetime":
            rollup_res = dim.rollup_resolution
            ref = [{"function": "rollup", "args": [ref, {"value": rollup_res}]}]
        elif dim.type == "categorical_array":
            ref = [{"each": dim.entity_url}, ref]
        elif dim.type == "multiple_response":
            ref = [{"each": dim.entity_url}, {"function": "as_selected", "args": [ref]}]
        else:
            ref = [ref]

        return ref


def count(*args):
    return {"function": "cube_count", "args": list(args)}



import mock
from unittest import TestCase

import requests

from pycrunch import elements, shoji

try:
    # Python 2
    from urllib import urlencode, quote
except ImportError:
    # Python 3
    from urllib.parse import urlencode, quote


class TestJSONObject(TestCase):

    class Foo(elements.JSONObject):
        pass

    def test_json_property(self):
        foo = self.Foo(bar=42)
        expected = '{\n    "bar": 42\n}'
        self.assertEqual(foo.pretty, expected)
        expected = '{"bar":42}'
        self.assertEqual(foo.json, expected)

    def test_str(self):
        foo = self.Foo(bar=42)
        expected = 'tests.test_elements.Foo(**{\n    "bar": 42\n})'
        self.assertEqual(str(foo), expected)

    def test_attribute_access(self):
        foo = self.Foo(bar=42)
        self.assertEqual(foo.bar, 42)

    def test_attribute_error(self):
        foo = self.Foo(bar=42)
        msg = 'Foo has no attribute nope'
        self.assertRaisesRegexp(AttributeError, msg, getattr, foo, 'nope')

    def test_copy(self):
        foo = self.Foo(bar=42)
        self.assertEqual(foo.copy(), {'bar': 42})


class TestElement(TestCase):

    class Foo(elements.Element):
        element = "shoji:foo"

    def test_copy(self):
        foo = self.Foo(session=None, bar=42)
        self.assertEqual(foo.copy(), {'bar': 42, 'element': 'shoji:foo'})


class TestParseElement(TestCase):

    class Shape(elements.Element):
        element = "shoji:shape"

    def test_unknown_element(self):
        # data dict should stay a dict because Color is unregistered
        data = {'name': 'blue', 'element': 'shoji:color'}
        result = elements.parse_element(session=None, j=data)
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result, data)

    def test_dict(self):
        # data dict should become a Shape instance
        data = {'name': 'circle', 'element': 'shoji:shape'}
        result = elements.parse_element(session=None, j=data)
        self.assertTrue(isinstance(result, self.Shape))
        self.assertEqual(result, data)

    def test_list_of_dicts(self):
        # each dict in the data list should become a Shape instance
        data = [
            {'name': 'circle', 'element': 'shoji:shape'},
            {'name': 'square', 'element': 'shoji:shape'},
        ]
        result = elements.parse_element(session=None, j=data)
        for element in result:
            self.assertTrue(isinstance(element, self.Shape))
        self.assertEqual(result, data)


class TestDocument(TestCase):

    class Person(elements.Document):
        element = "shoji:person"

    def _mkresp(self, **attrs):
        resp = requests.Response()
        for n in attrs:
            setattr(resp, n, attrs[n])
        return resp

    def _mk_follow_doc(self, catalog_url):
        session = mock.MagicMock()
        resp = shoji.Catalog(session, **{
            "element": "shoji:catalog",
            "self": catalog_url
        })
        session.get = mock.MagicMock(return_value=self._mkresp(payload=resp))

        entity = shoji.Entity(session, **{
            "element": "shoji:entity",
            "self": "self/url",
            'body': {},
            'catalogs': {
                'follow_me': catalog_url
            }
        })
        return entity

    def test_follow(self):
        catalog_url = '/catalog/url/'
        entity = self._mk_follow_doc(catalog_url)
        self.assertEqual(entity.follow("follow_me").self, catalog_url)

    def test_follow_qs_as_dict(self):
        catalog_url = '/catalog/url/'
        entity = self._mk_follow_doc(catalog_url)
        query = {
            "param1": "value1",
            "param2": "value2",
        }
        entity.follow("follow_me", query)
        entity.session.get.assert_called_with("%s?%s" % (catalog_url, urlencode(query)))

    def test_follow_qs_as_string(self):  # This is the old way of using .follow
        catalog_url = '/catalog/url/'
        entity = self._mk_follow_doc(catalog_url)

        # Note that the query is a urlencoded string
        query = urlencode({
            "param1": "value1",
            "param2": "value2",
        })
        entity.follow("follow_me", query)
        entity.session.get.assert_called_with("%s?%s" % (catalog_url, query))

    def test_follow_uri_template(self):
        catalog_url = '/catalog/{name}/'
        entity = self._mk_follow_doc(catalog_url)

        # Note that the query is a urlencoded string
        query = {
            "param1": "value1",
            "param2": "value2",
            "name": "replaced/evil name"
        }
        entity.follow("follow_me", query)
        call_url = "%s?%s" % (catalog_url.replace("{name}", quote(query['name'], safe="")), urlencode({
            "param1": "value1",
            "param2": "value2"
        }))
        entity.session.get.assert_called_with(call_url)

    def test_follow_no_link(self):
        person = self.Person(session=None, self='some uri')
        msg = 'Person has no link foo'
        self.assertRaisesRegexp(AttributeError, msg, person.follow, 'foo')

    def test_refresh(self):
        before = {
            'old_field': 'should remove',
            'self': 'some uri', 'element': 'shoji:person'
        }
        after = {
            'new_field': 'should exist',
            'self': 'some uri', 'element': 'shoji:person'
        }

        session_mock = mock.MagicMock()
        response = self._mkresp(payload=after)
        session_mock.get = mock.MagicMock(return_value=response)

        person = self.Person(
            session=session_mock,
            self='some uri',
            old_field='should remove'
        )
        self.assertEqual(person, before)

        person.refresh()
        self.assertEqual(person, after)
        session_mock.get.assert_called_once_with('some uri')

    def test_refresh_no_response(self):
        session_mock = mock.MagicMock()
        response = self._mkresp(payload=None)
        session_mock.get = mock.MagicMock(return_value=response)

        person = self.Person(session=session_mock, self='some uri')
        msg = 'Response could not be parsed.'
        self.assertRaisesRegexp(TypeError, msg, person.refresh)
        session_mock.get.assert_called_once_with('some uri')

    def test_post(self):
        session_mock = mock.MagicMock()
        person = self.Person(session=session_mock, self='some uri')
        person.post(data={'age': 10})
        session_mock.post.assert_called_once_with(
            'some uri',
            '{"age": 10}',
            headers={'Content-Type': 'application/json'}
        )

    def test_put(self):
        session_mock = mock.MagicMock()
        person = self.Person(session=session_mock, self='some uri')
        person.put(data={'age': 10})
        session_mock.put.assert_called_once_with(
            'some uri',
            '{"age": 10}',
            headers={'Content-Type': 'application/json'}
        )

    def test_patch(self):
        session_mock = mock.MagicMock()
        person = self.Person(session=session_mock, self='some uri')
        person.patch(data={'age': 10})
        session_mock.patch.assert_called_once_with(
            'some uri',
            '{"age": 10}',
            headers={'Content-Type': 'application/json'}
        )

    def test_delete(self):
        session_mock = mock.MagicMock()
        person = self.Person(session=session_mock, self='some uri')
        person.delete()
        session_mock.delete.assert_called_once_with('some uri')



# -*- coding: utf-8 -*-
import json

import mock
from six.moves.urllib_parse import urljoin
from unittest import TestCase

import sys
from requests import Response

from pycrunch.progress import DefaultProgressTracking, SimpleTextBarProgressTracking
from pycrunch.shoji import Catalog, CreateMixin, TaskProgressTimeoutError, TaskError, Index, Order, Entity
from pycrunch.lemonpy import URL


class TestShojiCreation(TestCase):
    def _mkresp(self, **attrs):
        resp = Response()
        for n in attrs:
            setattr(resp, n, attrs[n])
        return resp

    def test_create_does_post_catalog(self):
        sess = mock.MagicMock()
        c = Catalog(self='http://host.com/catalog', session=sess)
        c.create({'somedata': 1})
        sess.post.assert_called_once_with(
            'http://host.com/catalog',
            json.dumps({"somedata": 1, "body": {}, "element": "shoji:entity"}, indent=None, separators=(',', ':')),
            headers={'Content-Type': 'application/json'}
        )

    def test_create_does_post_entity(self):
        sess = mock.MagicMock()
        e = Entity(self='/entity/url/', session=sess)
        e.create({'somedata': 1})
        sess.post.assert_called_once_with(
            '/entity/url/',
            json.dumps({"somedata": 1, "body": {}, "element": "shoji:entity"}, indent=None, separators=(',', ':')),
            headers={'Content-Type': 'application/json'}
        )

    def test_create_waits_for_progress(self):
        sess = mock.MagicMock()
        sess.progress_tracking = DefaultProgressTracking(timeout=1.0, interval=0.1)
        sess.post = mock.MagicMock(return_value=self._mkresp(
            status_code=202,
            headers={'Location': 'http://host.com/somewhere'},
            payload={"value": 'http://host.com/progress/1'}
        ))
        sess.get = mock.MagicMock(side_effect=[
            self._mkresp(status_code=200, payload={'value': {'progress': 30}}),
            self._mkresp(status_code=200, payload={'value': {'progress': 60}}),
            self._mkresp(status_code=200, payload={'value': {'progress': 100}}),
        ])

        c = Catalog(self='http://host.com/catalog', session=sess)
        c.create({'somedata': 1})

        # Assert progress got called until it completed
        self.assertEqual(sess.get.call_count, 3)

    def test_create_timesout(self):
        sess = mock.MagicMock()
        sess.progress_tracking = DefaultProgressTracking(timeout=0.1, interval=0.1)
        sess.post = mock.MagicMock(return_value=self._mkresp(
            status_code=202,
            headers={'Location': 'http://host.com/somewhere'},
            payload={"value": 'http://host.com/progress/1'}
        ))
        sess.get = mock.MagicMock(side_effect=[
            self._mkresp(status_code=200, payload={'value': {'progress': 30}}),
            self._mkresp(status_code=200, payload={'value': {'progress': 60}}),
            self._mkresp(status_code=200, payload={'value': {'progress': 100}}),
        ])

        c = Catalog(self='http://host.com/catalog', session=sess)
        self.assertRaises(TaskProgressTimeoutError, c.create, {'somedata': 1})

    def test_create_timesout_continuation(self):
        sess = mock.MagicMock()
        sess.progress_tracking = DefaultProgressTracking(timeout=0.1, interval=0.1)
        sess.post = mock.MagicMock(return_value=self._mkresp(
            status_code=202,
            headers={'Location': 'http://host.com/somewhere'},
            payload={"value": 'http://host.com/progress/1'}
        ))
        sess.get = mock.MagicMock(side_effect=[
            self._mkresp(status_code=200, payload={'value': {'progress': 30}}),
            self._mkresp(status_code=200, payload={'value': {'progress': 60}}),
            self._mkresp(status_code=200, payload={'value': {'progress': 100}}),
        ])

        c = Catalog(self='http://host.com/catalog', session=sess)

        try:
            c.create({'somedata': 1})
        except TaskProgressTimeoutError as e:
            sess.progress_tracking = DefaultProgressTracking(timeout=None, interval=0.1)
            e.entity.wait_progress(e.response)
            self.assertEqual(sess.get.call_count, 3)
        else:
            assert False, "Should have raised TaskProgressTimeoutError"

    def test_create_raises_failures(self):
        sess = mock.MagicMock()
        sess.progress_tracking = DefaultProgressTracking(timeout=1.0, interval=0.1)
        sess.post = mock.MagicMock(return_value=self._mkresp(
            status_code=202,
            headers={'Location': 'http://host.com/somewhere'},
            payload={"value": 'http://host.com/progress/1'}
        ))
        sess.get = mock.MagicMock(side_effect=[
            self._mkresp(status_code=200, payload={'value': {'progress': -1,
                                                             'message': 'Some Failure'}}),
        ])

        c = Catalog(self='http://host.com/catalog', session=sess)
        self.assertRaises(TaskError, c.create, {'somedata': 1})

    def test_create_progressbar(self):
        sess = mock.MagicMock()
        sess.post = mock.MagicMock(return_value=self._mkresp(
            status_code=202,
            headers={'Location': 'http://host.com/somewhere'},
            payload={"value": 'http://host.com/progress/1'}
        ))
        sess.get = mock.MagicMock(side_effect=[
            self._mkresp(
                status_code=200,
                payload={'value': {'progress': i * 10}}
            ) for i in range(11)
        ])

        c = Catalog(self='http://host.com/catalog', session=sess)

        class FakeStdout(object):
            writes = []

            def write(self, text):
                self.writes.append(text)

            def flush(self):
                pass

        with mock.patch.object(sys, 'stdout', FakeStdout()):
            c.create(
                {'somedata': 1},
                progress_tracker=SimpleTextBarProgressTracking(
                    timeout=None, interval=0.1
                )
            )

        # Check we printed the progressbar up to 100%.
        self.assertEqual(''.join(FakeStdout.writes).count('-'),
                         SimpleTextBarProgressTracking.BAR_WIDTH)

    def test_accepts_document_instance(self):
        sess = mock.MagicMock()
        location = 'http://host.com/somewhere'
        sess.post = mock.MagicMock(return_value=self._mkresp(
            status_code=201,
            headers={'Location': location}
        ))
        catalog = Catalog(self='http://host.com/catalog', session=sess)
        body = {
            "name": 'subcatalog'
        }
        sub_catalog = Catalog(session=sess, body=body)
        retval = catalog.create(sub_catalog)
        self.assertEqual(retval.self, location)
        self.assertEqual(retval.body, body)


class TestIndex(TestCase):
    def test_relative_access(self):
        base_url = URL('http://host.name/base/url/', None)
        session = mock.MagicMock()

        rel_url = '../folder/item/'
        abs_url = URL(rel_url, base_url).absolute

        tup = {'id': 1}  # Some arbitrary value

        rel_index = Index(session, base_url, **{
            rel_url: tup
        })
        abs_index = Index(session, base_url, **{
            abs_url: tup
        })

        abs_on_rel = rel_index[abs_url]
        rel_on_abs = abs_index[rel_url]

        self.assertEqual(abs_on_rel, rel_on_abs)
        self.assertEqual(abs_on_rel, tup)
        self.assertEqual(rel_on_abs, tup)

    def test_catalog_follow_entity(self):
        base_url = URL('http://host.name/catalog/', None)

        ent_1_url = urljoin(base_url, '01/')
        ent_2_url = urljoin(base_url, '02/')
        ent_3_url = urljoin(base_url, '03/')
        ent_4_url = urljoin(base_url, '04/')
        entities = {
            ent_1_url: {
                'full_entity': True,
                'name': 'Ent 01'
            },
            ent_2_url: {
                'full_entity': True,
                'name': 'Ent 02'
            },
            ent_3_url: {
                'full_entity': True,
                'name': 'Ent 03'
            },
            ent_4_url: {
                'full_entity': True,
                'name': 'Ent 04'
            }
        }

        fetches = []

        def _get(ent_url, *args, **kwargs):
            fetches.append((ent_url, args, kwargs))
            resp = mock.Mock(payload=entities[ent_url])
            return resp

        session = mock.MagicMock()
        session.get = _get
        payload = {
            'element': 'shoji:catalog',
            'self': base_url,
            # This index has mixed urls as kes
            'index': {
                './01/': {
                    'name': '01'
                },
                ent_2_url: {
                    'name': '02'
                }
            }
        }

        index = Index(session, base_url, **payload['index'])
        # Continue growing the index using other ways, __setitem__ and .update
        index['03/'] = {  # This has another way of being valid relative
            'name': '03'
        }
        index.update({
            '../catalog/04/': {  # Yet way of being valid relative
                'name': '04'
            }
        })

        self.assertTrue(index[ent_1_url].entity['full_entity'])
        self.assertEqual(index[ent_1_url].entity['name'], 'Ent 01')
        self.assertTrue(index['01/'].entity['full_entity'])
        self.assertEqual(index['01/'].entity['name'], 'Ent 01')
        self.assertTrue(index['./01/'].entity['full_entity'])
        self.assertEqual(index['./01/'].entity['name'], 'Ent 01')
        self.assertTrue(index['../catalog/01/'].entity['full_entity'])
        self.assertEqual(index['../catalog/01/'].entity['name'], 'Ent 01')

        self.assertTrue(index[ent_2_url].entity['full_entity'])
        self.assertEqual(index[ent_2_url].entity['name'], 'Ent 02')
        self.assertTrue(index['02/'].entity['full_entity'])
        self.assertEqual(index['02/'].entity['name'], 'Ent 02')
        self.assertTrue(index['./02/'].entity['full_entity'])
        self.assertEqual(index['./02/'].entity['name'], 'Ent 02')
        self.assertTrue(index['../catalog/02/'].entity['full_entity'])
        self.assertEqual(index['../catalog/02/'].entity['name'], 'Ent 02')

        self.assertTrue(index[ent_3_url].entity['full_entity'])
        self.assertEqual(index[ent_3_url].entity['name'], 'Ent 03')
        self.assertTrue(index['03/'].entity['full_entity'])
        self.assertEqual(index['03/'].entity['name'], 'Ent 03')
        self.assertTrue(index['./03/'].entity['full_entity'])
        self.assertEqual(index['./03/'].entity['name'], 'Ent 03')
        self.assertTrue(index['../catalog/03/'].entity['full_entity'])
        self.assertEqual(index['../catalog/03/'].entity['name'], 'Ent 03')

        self.assertTrue(index[ent_4_url].entity['full_entity'])
        self.assertEqual(index[ent_4_url].entity['name'], 'Ent 04')
        self.assertTrue(index['04/'].entity['full_entity'])
        self.assertEqual(index['04/'].entity['name'], 'Ent 04')
        self.assertTrue(index['./04/'].entity['full_entity'])
        self.assertEqual(index['./04/'].entity['name'], 'Ent 04')
        self.assertTrue(index['../catalog/04/'].entity['full_entity'])
        self.assertEqual(index['../catalog/04/'].entity['name'], 'Ent 04')

        self.assertEqual(fetches, [
            (ent_1_url, (), {}),
            (ent_2_url, (), {}),
            (ent_3_url, (), {}),
            (ent_4_url, (), {}),
        ])


class TestOrders(TestCase):
    def test_follows_catalogs(self):
        catal_url = '/catalog/url/'
        session = mock.Mock(**{
            'get': lambda x: mock.Mock(**{'payload.self': catal_url})
        })

        order = Order(session, **{
            'graph': [],
            'catalogs': {
                'follow_me': catal_url
            }
        })
        self.assertEqual(order.follow_me.self, catal_url)


class TestEntities(TestCase):
    def test_entities_can_have_index(self):
        ent_url = '/entity/url/'
        session = mock.Mock()
        body = {
            'attr': 'val'
        }
        index = {
            'url1/': {'key': 'val1'},
            'url2/': {'key': 'val2'}
        }
        ent = Entity(session, **{
            'self': ent_url,
            'body': body,
            'index': index
        })
        self.assertTrue(isinstance(ent.index, Index))
        self.assertEqual(ent.by('key')['val1'].entity_url, 'url1/')


class TestCreateMixin(TestCase):

    def test_by_integer_keys(self):
        tuple1 = {'id': 1, 'name': 'A'}
        tuple2 = {'id': 2, 'name': 'B'}

        obj = CreateMixin()
        obj.index = {'url1/': tuple1, 'url2/': tuple2}

        self.assertEqual(obj.by('id'), {1: tuple1, 2: tuple2})



from pycrunch import elements


def cast(variable, type, format=None, offset=None, resolution=None):
    """Alter the given variable Entity to a new type.

    Various parameters may need to be sent to properly convert from one type
    to another. Datetime is particularly demanding.
    """
    payload = elements.JSONObject(cast_as=type)
    if format is not None:
        payload['format'] = format
    if offset is not None:
        payload['offset'] = offset
    if resolution is not None:
        payload['resolution'] = resolution

    return variable.cast.post(data=payload.json)



from mock import Mock, patch
import pytest

from pycrunch.cubes import DimensionsPreparer


def test_prepare_ref(prepare_ref_fixture):
    dim, expected = prepare_ref_fixture
    actual = DimensionsPreparer.prepare_ref(dim)
    assert actual == expected


def test_get_dimension_by_string(dimension_string_fixture):
    dim_str, preparer, dim = dimension_string_fixture
    assert preparer.get_dimension_by_string(dim_str) == dim


# fixtures -----------------------------------------------------------


@pytest.fixture(params=[
    ('categorical', None, [{'variable': 'fake_url'}]),
    ('numeric', None, [{'function': 'bin', 'args': [{'variable': 'fake_url'}]}]),
    ('datetime', 'fake_rollup', [{'function': 'rollup', 'args': [{'variable': 'fake_url'}, {'value': 'fake_rollup'}]}]),
    ('datetime', None, [{'function': 'rollup', 'args': [{'variable': 'fake_url'}, {'value': None}]}]),
    ('multiple_response', None, [{'each': 'fake_url'}, {'function': 'as_selected', 'args': [{'variable': 'fake_url'}]}]),
])
def prepare_ref_fixture(request):
    dimension_type, rollup, expected = request.param
    dim = Mock(entity_url='fake_url')
    dim.type = dimension_type
    dim.rollup_resolution = rollup
    return dim, expected

@pytest.fixture(params=[
    ('fake_url', 'URL', Mock(), None),
    ('fake_url/subvariables/fake_id', 'URL_SUBVAR', Mock(), Mock()),
    ('fake_alias', 'ALIAS', Mock(), None),
    ('fake_alias', 'NAME', Mock(), None),
])
def dimension_string_fixture(request):
    dim_str, str_type, dim, subvar = request.param
    preparer = DimensionsPreparer(Mock())
    preparer._dataset.variables.index = dict()
    preparer._variables_by_alias = dict()
    preparer._variables_by_name = dict()
    if str_type == 'URL':
        preparer._dataset.variables.index[dim_str] = dim
    elif str_type == 'URL_SUBVAR':
        var_url = dim_str.split('subvariables/')[0]
        preparer._dataset.variables.index[var_url] = dim
        preparer._dataset.variables.index[
            var_url
        ].entity.subvariables.index = {dim_str: subvar}
        dim = subvar
    elif str_type == 'ALIAS':
        preparer._variables_by_alias[dim_str] = dim
    elif str_type == 'NAME':
        preparer._variables_by_name[dim_str] = dim
    return dim_str, preparer, dim



"""PyCrunch

A Python client library for Crunch.io.


Using pycrunch
--------------

To use pycrunch in your project, run:

    $ python setup.py develop

This will make the code in this directory available to other projects.

Getting started
---------------

Start a simple site session via:

    >> import pycrunch
    >> site = pycrunch.connect("me@mycompany.com", "yourpassword", "https://app.crunch.io/api/")

Or, if you have a crunch access token:

    >> import pycrunch
    >> site = pycrunch.connect_with_token("DFIJFIJWIEJIJFKSJLKKDJKFJSLLSLSL", "https://app.crunch.io/api/")


Then, you can browse the site. Use `print` to pretty-indent JSON payloads:

    >> print site
    pycrunch.shoji.Catalog(**{
        "element": "shoji:catalog",
        "self": "https://app.crunch.io/api/",
        "description": "The API root.",
        "catalogs": {
            "datasets": "https://app.crunch.io/api/datasets/",
            ...
        },
        "urls": {
            "logout_url": "https://app.crunch.io/api/logout/",
            ...
        },
        "views": {
            "migration": "https://app.crunch.io/api/migration/"
        }
    })

URI's in payloads' catalogs, views, fragments, and urls collections
are followable automatically:

    >> print site.datasets
    pycrunch.shoji.Catalog(**{
        "self": "https://app.crunch.io/api/datasets/",
        "element": "shoji:catalog",
        "index": {
            "https://app.crunch.io/api/datasets/dbf9fca7b727/": {
                "owner_display_name": "me@mycompany.com",
                "description": "",
                "id": "dbf9fca7b727",
                "owner_id": "https://app.crunch.io/api/users/253b68/",
                "archived": false,
                "name": "Hog futures tracking (May 2014)"
            },
        },
        ...
    })

Each recognized JSON payload also automatically gives dotted-attribute
access to the members of each JSON object:

    >> print site.datasets.index.values()[0]
    pycrunch.shoji.Tuple(**{
        "size": {
            "rows": 6,
            "columns": 8
        },
        "archived": false,
        "owner_name": "Me",
        "description": "",
        "end_date": null,
        "owner_id": "https://app.crunch.io/api/users/a69402/",
        "current_editor": "https://app.crunch.io/api/users/a69402/",
        "creation_time": "2016-03-02T22:19:14.463000+00:00",
        "current_editor_name": "The Editor",
        "start_date": null,
        "permissions": {
            "edit": true,
            "change_permissions": true,
            "add_users": true,
            "change_weight": true,
            "view": true
        },
        "id": "1234",
        "name": "Some Kind of Love"
    })


Responses may also possess additional helpers, like the `entity` property of
each Tuple in a catalog's index, which follows the link to the Entity resource:

    >> print site.datasets.index.values()[0].entity_url
    "https://app.crunch.io/api/datasets/1234/"

    >> print site.datasets.index.values()[0].entity
    pycrunch.shoji.Entity(**{u'body': {u'archived': False,
           u'creation_time': u'2015-03-18T01:56:12.462000',
           u'current_editor': u'https://app.crunch.io/api/users/00002/',
           u'current_editor_name': u'Jean-Luc Picard',
           u'description': u'',
           u'end_date': None,
           u'id': u'1234',
           u'name': u'a_econ_few_columns',
           u'notes': u'',
           u'owner': u'https://app.crunch.io/api/users/00002/',
           u'permissions': {u'add_users': True,
                            u'change_permissions': True,
                            u'change_weight': True,
                            u'edit': True,
                            u'view': True},
           u'size': {u'columns': 14, u'rows': 1000},
           u'start_date': None,
           u'weight': u'https://app.crunch.io/api/datasets/1234/variables/c1820eb7befd4704beacfdbcb430969c/'},
           u'catalogs': {u'actions': u'https://app.crunch.io/api/datasets/1234/actions/',
                         u'batches': u'https://app.crunch.io/api/datasets/1234/batches/',
                         u'comparisons': u'https://app.crunch.io/api/datasets/1234/comparisons/',
                         u'decks': u'https://app.crunch.io/api/datasets/1234/decks/',
                         u'filters': u'https://app.crunch.io/api/datasets/1234/filters/',
                         u'forks': u'https://app.crunch.io/api/datasets/1234/forks/',
                         u'joins': u'https://app.crunch.io/api/datasets/1234/joins/',
                         u'multitables': u'https://app.crunch.io/api/datasets/1234/multitables/',
                         u'permissions': u'https://app.crunch.io/api/datasets/1234/permissions/',
                         u'savepoints': u'https://app.crunch.io/api/datasets/1234/savepoints/',
                         u'variables': u'https://app.crunch.io/api/datasets/1234/variables/',
                         u'weight_variables': u'https://app.crunch.io/api/datasets/1234/weight_variables/'},
           u'description': u'Detail for a given dataset',
           u'element': u'shoji:entity',
           u'fragments': {u'exclusion': u'https://app.crunch.io/api/datasets/1234/exclusion/',
                          u'main_deck': u'https://app.crunch.io/api/datasets/1234/decks/bf57e39d42ae472b91e03de9a299e7c4/',
                          u'state': u'https://app.crunch.io/api/datasets/1234/state/',
                          u'stream': u'https://app.crunch.io/api/datasets/1234/stream/',
                          u'table': u'https://app.crunch.io/api/datasets/1234/table/'},
           u'self': u'https://app.crunch.io/api/datasets/1234/',
           u'urls': {u'editor_url': u'https://app.crunch.io/api/users/00002/',
                     u'owner_url': u'https://app.crunch.io/api/users/00002/',
                     u'user_url': u'https://app.crunch.io/api/users/00002/'},
           u'views': {u'applied_filters': u'https://app.crunch.io/api/datasets/1234/filters/applied/',
                      u'cube': u'https://app.crunch.io/api/datasets/1234/cube/',
                      u'export': u'https://app.crunch.io/api/datasets/1234/export/',
                      u'summary': u'https://app.crunch.io/api/datasets/1234/summary/'}}
            )

You typically add new resources to a Catalog via its `create` method:

    >> ds = site.datasets.create({"body": {
            'name': "My first dataset"
        }}, refresh=True)
    >> gender = ds.variables.create({"body": {
            'name': 'Gender',
            'alias': 'gender',
            'type': 'categorical',
            'categories': [
                {'id': -1, 'name': 'No Data', 'numeric_value': None, 'missing': True},
                {'id': 1, 'name': 'M', 'numeric_value': None, 'missing': False},
                {'id': 2, 'name': 'F', 'numeric_value': None, 'missing': False}
            ],
            'values': [1, 2, {"?": -1}, 2]
        }}, refresh=True)
    >> print ds.table.data
    pycrunch.elements.JSONObject(**{
        "e7f361628": [
            1,
            2,
            {"?": -1},
            2
        ]
    })
"""

from six.moves import urllib

from pycrunch import cubes
from pycrunch import elements
from pycrunch import shoji
from pycrunch.shoji import TaskError, TaskProgressTimeoutError
from pycrunch import importing
from pycrunch.lemonpy import ClientError, ServerError, urljoin
from pycrunch.version import __version__

Session = elements.ElementSession

__all__ = [
    'cubes',
    'elements',
    'shoji',
    'TaskError', 'TaskProgressTimeoutError',
    'importing',
    'ClientError', 'ServerError', 'CrunchError'
    'Session',
    'urljoin',
    'connect', 'connect_with_token'
]


class CrunchError(elements.Element):

    element = "crunch:error"


class CrunchTable(elements.Document):

    element = "crunch:table"


session = None


def connect(user, pw, site_url="https://app.crunch.io/api/",
            progress_tracking=None, session_class=Session):
    """
    Log in to Crunch with a user/pw; return the top-level Site payload.  Using
    this or the other connect method (the first time only) stores a reference
    to the session created in pycrunch.session for future use.

    Returns the API Root Entity, or errors if unable to connect.
    """
    global session
    ret = session_class(
        user, pw, progress_tracking=progress_tracking
    ).get(site_url).payload
    if session is None:
        session = ret
    return ret


def connect_with_token(token, site_url="https://us.crunch.io/api/",
                       progress_tracking=None, session_class=Session):
    """
    Log in to Crunch with a token; return the top-level Site payload. Using
    this or the other connect method (the first time only) stores a reference
    to the session created in pycrunch.session for future use.

    Returns the API Root Entity, or errors if unable to connect.
    """
    global session
    ret = session_class(
        token=token,
        domain=urllib.parse.urlparse(site_url).netloc,
        progress_tracking=progress_tracking
    ).get(site_url).payload
    if session is None:
        session = ret
    return ret


def get_dataset(dataset_name_or_id, site=None):
    """
    Retrieve a reference to a given dataset (either by name, or ID) if it exists.
    This method uses the library singleton session if the optional "site"
    parameter is not provided.

    Returns a Dataset Entity record if the dataset exsists.
    Raises a KeyError if no such dataset exists.
    """
    global session
    if site is None:
        site = session

    ds_catalog = site.datasets
    try:
        dataset = ds_catalog.by('name')[dataset_name_or_id].entity
    except KeyError:
        dataset = ds_catalog.by('id')[dataset_name_or_id].entity
    return dataset



"""IPython importable module."""

from pycrunch import pandaslib


def load_ipython_extension(ipython):
    # TODO: auto-inject the site in builtins drawn from an env var
    # token = ipython.config['InteractiveShellApp'].crunch_token
    # or
    # token = os.environ.get("CRUNCH_TOKEN", None)
    dataframe = pandaslib.dataframe_from_dataset
    ipython.ns_table["builtin"]["dataframe_from_dataset"] = dataframe

    # TODO: do we want any magic?
    # ipython.register_magic_function(
    #     lambda line: pandaslib.dataframe_from_dataset(line.split(" ")),
    #     magic_kind='line', magic_name="crunch"
    # )
    pass


def unload_ipython_extension(ipython):
    # If you want your extension to be unloadable, put that logic here.
    pass



# -*- coding: utf-8 -*-
import sys

DEFAULT_PROGRESS_TIMEOUT = 30
DEFAULT_PROGRESS_INTERVAL = 0.5


class _ProgressTrackingConfig(object):
    """Base class for all progress tracking configurations."""
    def __init__(self, timeout, interval):
        self.timeout = timeout
        self.interval = interval

    def start_progress(self):
        raise NotImplementedError

    def on_progress(self, state, progress):
        raise NotImplementedError


class DefaultProgressTracking(_ProgressTrackingConfig):
    """Default progress tracking configuration.

    Waits for 30 seconds, polls every 0.5 seconds and
    does not report progress in any way.
    """
    def __init__(self, timeout=DEFAULT_PROGRESS_TIMEOUT,
                 interval=DEFAULT_PROGRESS_INTERVAL):
        super(DefaultProgressTracking, self).__init__(timeout, interval)

    def start_progress(self):
        return None

    def on_progress(self, state, progress):
        pass


class SimpleTextBarProgressTracking(DefaultProgressTracking):
    """Progress tracking config that displays a progress bar on stdout.

    Behaves like `DefaultProgressTracking` but provides a very simple
    textual progress bar, its purpose is mostly to showcase how a
    custom progress tracker can be implemented.
    """
    BAR_WIDTH = 40

    def start_progress(self):
        sys.stdout.write("[%s]" % (" " * self.BAR_WIDTH))
        sys.stdout.flush()
        sys.stdout.write("\b" * (self.BAR_WIDTH + 1))
        return dict(count=0)

    def on_progress(self, state, progress):
        if progress['progress'] in (-1, 0):
            return

        step = self.BAR_WIDTH / 100.0
        progress_value = step * progress['progress']
        while state['count'] < progress_value:
            sys.stdout.write("-")
            state['count'] += 1
        sys.stdout.flush()

        if progress['progress'] == 100:
            sys.stdout.write("\n")



from __future__ import division

import logging

import six
from six.moves import urllib
from six.moves.http_cookiejar import Cookie

import requests

requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)
urljoin = requests.compat.urljoin


class LemonPyError(Exception):
    """Base class for Exceptions which occur within LemonPy."""
    pass


class ClientError(LemonPyError):

    def __init__(self, response):
        args = []
        if not isinstance(response, six.string_types):
            args = [response.request.url, response.payload]

        super(ClientError, self).__init__(response, *args)

    @property
    def status_code(self):
        return self.args[0].status_code

    @property
    def message(self):
        return self.args[2]['message']


class ServerError(LemonPyError):

    def __init__(self, response):
        args = []
        if not isinstance(response, six.string_types):
            args = [response.request.url, response.payload]
        super(ServerError, self).__init__(response, *args)


class ResponseHandler(object):
    """A requests.Session response-hook aware of status and Content-Type.

    When an instance of this class is set as Session.hooks['response'],
    all Response objects are passed to its __call__ method. That method
    looks for a handler on itself for the specific response status code,
    such as "self.status_204(r)", passes the Response object to it,
    and returns its output (which MUST be a Response object, and is
    typically the given one). If no specific handler method is found,
    a more general handler is sought for the class of the response status,
    such as "self.status_2xx(r)". If neither is found, an AttributeError
    is raised.

    Assuming a handler is found, it should call self.parse_payload(r),
    which attempts to parse the Response based on its Content-Type.
    """

    parsers = {
        'application/json': lambda session, r: r.json()
    }

    def __init__(self, session):
        self.session = session

    def __call__(self, r, *args, **kwargs):
        code = r.status_code

        # First, try the specific response status code
        handler = getattr(self, "status_%d" % code, None)
        if handler is not None:
            return handler(r)

        # If that fails, look for a Nxx code
        handler = getattr(self, "status_%dxx" % (code // 100), None)
        if handler is not None:
            return handler(r)

        raise AttributeError("No handler found for response status %s" %
                             r.status_code)

    def parse_payload(self, r):
        """Attach a .payload to the given Response.

        If the Content-Type of the given Response has a parser function
        registered in self.parsers, it will be called as parser(session, r);
        whatever it returns will be attached as r.payload. If no parser
        function exists, r.payload is set to None, and the caller will
        have to examine the Response directly to determine its payload.
        """
        ct = r.headers.get("Content-Type", "").split(";", 1)[0]
        parser = self.parsers.get(ct)
        r.payload = parser(self.session, r) if parser else None

    def status_2xx(self, r):
        self.parse_payload(r)
        return r

    # status_201 = status_2xx # The caller must look up r.headers["Location"]

    def status_204(self, r):
        # TODO: should we warn here if there is a payload?
        r.payload = None
        return r

    def status_301(self, r):
        # Support permanent redirects
        return r

    def status_302(self, r):
        # Support redirects
        return r

    def status_303(self, r):
        self.parse_payload(r)
        return r

    def status_4xx(self, r):
        self.parse_payload(r)
        raise ClientError(r)

    def status_5xx(self, r):
        self.parse_payload(r)
        raise ServerError(r)


def make_cookie(name, value, domain):
    '''
        Makes a cookie with provided name and value.
    '''
    return Cookie(
        version=0,
        name=name,
        value=value,
        port=None,
        port_specified=False,
        domain=domain,
        domain_specified=True,
        domain_initial_dot=False,
        path="/",
        path_specified=True,
        secure=False,
        expires=None,
        discard=False,
        comment=None,
        comment_url=None,
        rest=None
    )


class Session(requests.Session):

    headers = {
        "Accept-Encoding": "gzip",
    }
    handler_class = ResponseHandler

    def __init__(self):
        super(Session, self).__init__()

        self.headers.update(self.__class__.headers)
        self.proxies = urllib.request.getproxies()

        if self.token:
            domain = self.domain or 'local.crunch.io'
            self.cookies.set_cookie(make_cookie('token', self.token, domain))

        self.hooks["response"] = self.handler_class(self)


class URL(str):
    """A subclass of str for URL's. self.absolute = urljoin(self.base, self)."""

    def __new__(cls, value, *args, **kwargs):
        return str.__new__(cls, value)

    def __init__(self, value, base):
        self.base = base

    @property
    def absolute(self):
        """Return self, which may be relative to self.base, as an absolute URL."""
        return urljoin(self.base, self)

    def relative_to(self, base):
        """Return self, relative to the given absolute base."""
        base = urllib.parse.urlparse(base)
        new = urllib.parse.urlparse(self.absolute)

        if base.scheme != new.scheme or base.netloc != new.netloc:
            return self.absolute

        base_path = base.path.split('/')[:-1]
        new_path = new.path.split('/')
        while base_path and new_path:
            a, b = base_path[0], new_path[0]
            if a != b:
                break
            base_path.pop(0)
            new_path.pop(0)
        new_path = (['..'] * len(base_path)) + new_path
        new_path = '/'.join(new_path)

        return urllib.parse.urlunparse(
            ("", "", new_path, new.params, new.query, new.fragment)
        )



import six
from pandas import DataFrame, Categorical, Series, to_datetime


def series_from_variable(col, vardef):
    """Return the given Crunch column and variable def as a Pandas Series."""
    col = [None if (isinstance(item, dict) and list(item.keys()) == ['?']) else item
           for item in col]

    if vardef.type == 'categorical':
        cats = dict((c['id'], c['name']) for c in vardef['categories'])
        col = [None if val is None else cats[val] for val in col]
        return Categorical(
            col,
            categories=[cat['name'] for cat in vardef['categories']
                        if not cat['missing']],
            ordered=True
        )
    elif vardef.type == 'datetime':
        return to_datetime(Series(col))

    return Series(col)


ROWCHUNKSIZE = 1000


def dataframe(dataset, variables=None):
    """Return a Pandas DataFrame for the given Crunch Dataset Entity object.
    Retrieve a dataset using pycrunch.get_dataset("dataset name or id").

    If the 'variables' argument is given and not None, it should be a
    single variable alias, or a list of such, in which case only those
    variables will be included in the DataFrame. If omitted or None,
    all variables are included.

    The returned DataFrame has an extra "metadata" attribute on it:
    a dict of Crunch variable definitions for each Series (keyed by id).
    """
    data = {}

    if variables is None:
        numrows = dataset.summary.value.unweighted.total
        seenrows = 0
        all_data = {}
        while True:
            t = dataset.session.get(
                "%s?offset=%d&limit=%d" %
                (dataset.fragments['table'], seenrows, ROWCHUNKSIZE)
            ).payload
            for name, value in six.iteritems(t.data):
                if name not in all_data:
                    all_data[name] = []
                all_data[name].extend(value)
            seenrows += ROWCHUNKSIZE
            if seenrows > numrows:
                break

        metadata = t.metadata

        # Convert to Series
        for varid, col in six.iteritems(all_data):
            vardef = t.metadata[varid]
            data[vardef.alias] = series_from_variable(col, vardef)
    else:
        metadata = {}
        if not isinstance(variables, list):
            variables = list(variables)

        var_catalog = dataset.variables
        for variable in variables:
            try:
                t = var_catalog.by('alias')[variable].entity
            except KeyError:
                raise KeyError('No variable with alias: %s' % variable)
            vardef = t.body
            value = dataset.session.get(t.views['values']).payload['value']
            data[vardef.alias] = series_from_variable(value, vardef)
            metadata[vardef.id] = vardef

    df = DataFrame(data)

    # Attach the crunch:table metadata object to the df
    # so consumers can have both.
    df.metadata = metadata

    return df



# coding: utf-8
import json
from .lemonpy import URL
from .shoji import wait_progress


def export_dataset(dataset, options, format='csv', progress_tracker=None):
    """
    Exports a Crunch dataset in the desired format. This is a blocking function
    call that will return a url where to download the exported file from.

    :param dataset: Shoji entity pointing to a Crnuch dataset
    :param options: Dictionary containing exporting options (where, filter, etc)
    :param format: Export format, CSV or SPSS
    :return: URL instance containing the url for the final file download
    """
    session = dataset.session
    endpoint = dataset.export.views[format]
    r = session.post(endpoint, json.dumps(options))
    dest_file = URL(r.headers['Location'], '')
    if r.status_code == 202:
        try:
            progress_url = r.payload['value']  # noqa
        except Exception:
            # Not a progress API just return the incomplete entity.
            # User will refresh it.
            pass
        else:
            # We have a progress_url, wait for completion
            wait_progress(r, session, progress_tracker)
    return dest_file



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from unittest import TestCase

import six

from pycrunch import csvlib


class TestCSV(TestCase):

    def test_unicode_values(self):
        # CSV rendering should handle simple unicode
        rows = [['☃']]
        csvlib.rows_as_csv_file(rows)

    def test_result_is_binary(self):
        # Result should be a stream with a binary type
        res = csvlib.rows_as_csv_file([['foo']])
        assert isinstance(next(res), six.binary_type)

    def test_rows_as_csv_file(self):
        # None should be emitted as an empty cell (unquoted)
        rows = [[0, 1, None, "bananas"]]
        fp = csvlib.rows_as_csv_file(rows)
        self.assertEqual(fp.read(), b'0,1,,"bananas"\n')

    def test_rows_as_csv_file_clean(self):
        # None should be emitted as an empty string (quoted)
        rows = [[0, 1, None, "bananas"]]
        fp = csvlib.rows_as_csv_file_clean(rows)
        self.assertEqual(fp.read(), b'0,1,"","bananas"\n')



"""LemonPy handlers for JSON payloads, especially those with an "element" member.

The base lemonpy parser for JSON returns the JSON text deserialized into
Python dicts, lists, etc. This module does the same, but wherever the base
parser would return a dict, it replaces that with a JSONObject instance.
JSONObject is itself a subclass of dict, so all the dict attributes and
methods are still present; however, the JSONObject has a .json property
and a nicer str output. It also allows reading keys as if they were
attributes, so instead of foo['bar'][0]['baz'] you can write foo.bar[0].baz.

When a JSON object has an "element" member, however, a specialized subclass
of JSONObject is chosen instead. These include a reference to the session,
and can therefore implement additional HTTP requests very naturally.
Any subclass of Element is automatically registered using its .element
class attribute. For example, the class definition:

    class Foo(Element):
        element = "myapp:foo"

...would result in any JSON object with an {"element": "myapp:foo"} member
being parsed into an instance of Foo, rather than a bare JSONObject.
"""

import json

import six

from requests.utils import get_environ_proxies

from pycrunch import lemonpy
from pycrunch.progress import DefaultProgressTracking
from .version import __version__

try:
    # Python 2
    from urllib import urlencode, quote
except ImportError:
    # Python 3
    from urllib.parse import urlencode, quote

omitted = object()


class JSONObject(dict):
    """A base class for JSON objects."""

    @property
    def json(self):
        return json.dumps(self, indent=None, separators=(',', ':'))

    @property
    def pretty(self):
        return json.dumps(self, indent=4)

    def __str__(self):
        return "%s.%s(**%s)" % (self.__module__, self.__class__.__name__,
                                # __str__ is indented!
                                json.dumps(self, indent=4))

    def __getattr__(self, key):
        # Return the requested attribute if present in self.keys
        v = self.get(key, omitted)
        if v is not omitted:
            return v

        raise AttributeError(
            "%s has no attribute %s" % (self.__class__.__name__, key))

    def copy(self):
        """Return a (shallow) copy of self."""
        return self.__class__(**self)


class ElementMeta(type):
    """A metaclass for Element subclasses which registers them."""

    def __new__(cls, name, bases, d):
        new_type = type.__new__(cls, name, bases, d)
        if 'element' in d:
            # Skip any subclass which has no 'element' attribute,
            # (including Element itself!) to allow for additional
            # base classes.
            elements[d['element']] = new_type
        return new_type


@six.add_metaclass(ElementMeta)
class Element(JSONObject):
    """A base class for JSON objects classified by an 'element' member."""

    def __init__(__this__, session, **members):
        __this__.session = session
        members.setdefault("element", __this__.__class__.element)
        super(Element, __this__).__init__(**members)

    def copy(self):
        """Return a (shallow) copy of self."""
        return self.__class__(self.session, **self)


elements = {}


def parse_element(session, j):
    """Recursively replace dict with appropriate subclasses of JSONObjects."""
    t = type(j)
    if t is dict:
        for k, v in list(six.iteritems(j)):
            j[k] = parse_element(session, v)

        elem = j.get("element", None)
        if elem in elements:
            return elements[elem](session, **j)
        else:
            return JSONObject(**j)
    elif t is list:
        return [parse_element(session, i) for i in j]
    else:
        return j


class Document(Element):
    """A base class for complete Documents classified by 'element'.

    In addition to normal __getattr__ access to members, this class
    adds automatic GETs to collections nominated as URL's.

    Documents are considered "top-level" objects that may be returned
    as a complete payload; they therefore include helper functions
    for refreshing themselves (via HTTP GET), plus post, put, patch,
    and delete. Not every resource is guaranteed to respond to all.
    """

    navigation_collections = ()

    def __getattr__(self, key):
        # Return the requested attribute if present in self.keys
        v = self.get(key, omitted)
        if v is not omitted:
            return v

        # If the requested attribute is present in a URL collection,
        # do a GET and return its payload.
        for collname in self.navigation_collections:
            coll = self.get(collname, {})
            if key in coll:
                url = coll[key]
                return self.session.get(url).payload

        raise AttributeError(
            "%s has no attribute %s" % (self.__class__.__name__, key))

    def follow(self, key, qs=None):
        """GET the payload of the requested collection URL."""
        url = None
        for collname in self.navigation_collections:
            coll = self.get(collname, {})
            if key in coll:
                url = coll[key]
                break

        if url is None:
            raise AttributeError(
                "%s has no link %s" % (self.__class__.__name__, key))

        if isinstance(qs, dict):
            template_lookups = {"{%s}" % k: k for k in qs.keys()}
            pruned_qs = {}
            for tpl, v in template_lookups.items():
                if tpl in url:
                    url = url.replace(tpl, quote(qs[v], safe=""))
                else:
                    pruned_qs[v] = qs[v]
            qs = urlencode(pruned_qs)

        if qs is not None:
            # Remove any existing qs, such as for URI Templates.
            url = url.rsplit("?", 1)[0] + "?" + qs
        return self.session.get(url).payload

    def refresh(self):
        """GET self.self, update self with its payload and return self."""
        r = self.session.get(self.self)
        if r.payload is None:
            raise TypeError("Response could not be parsed.", r)

        self.clear()
        self.update(r.payload)
        return self

    def post(self, data, *args, **kwargs):
        kwargs.setdefault('headers', {})
        kwargs["headers"].setdefault("Content-Type", "application/json")
        if not isinstance(data, six.string_types):
            data = json.dumps(data)
        return self.session.post(self.self, data, *args, **kwargs)

    def put(self, data, *args, **kwargs):
        kwargs.setdefault('headers', {})
        kwargs["headers"].setdefault("Content-Type", "application/json")
        if not isinstance(data, six.string_types):
            data = json.dumps(data)
        return self.session.put(self.self, data, *args, **kwargs)

    def patch(self, data, *args, **kwargs):
        kwargs.setdefault('headers', {})
        kwargs["headers"].setdefault("Content-Type", "application/json")
        if not isinstance(data, six.string_types):
            data = json.dumps(data)
        return self.session.patch(self.self, data, *args, **kwargs)

    def delete(self):
        return self.session.delete(self.self)

# -------------------------- HTTP request helpers -------------------------- #


def parse_json_element_from_response(session, r):
    """Return the appropriate Element instance if possible, otherwise JSON."""
    if not r.text:
        return JSONObject()
    j = r.json()

    return parse_element(session, j)


class ElementResponseHandler(lemonpy.ResponseHandler):
    """A lemonpy response handler which parses to JSONObjects and Elements.

    In addition, this subclass traps 401 Unauthorized responses,
    then attempts to authenticate to Crunch.io, and then repeats
    the request. That should probably be moved out somewhere else.
    """

    parsers = {
        'application/json': parse_json_element_from_response
    }

    def status_401(self, r):
        login_url = r.json()["urls"]["login_url"]
        if r.request.url == login_url:
            raise ValueError("Log in was not successful.")

        creds = {'email': self.session.email, 'password': self.session.password}
        login_r = self.session.post(
            login_url,
            headers={"Content-Type": "application/json"},
            data=json.dumps(creds)
        )

        # Repeat the request now that we've logged in. What a hack.
        r.request.headers['Cookie'] = login_r.headers['Set-Cookie']
        env_proxies = get_environ_proxies(r.request.url, no_proxy=None)
        r2 = self.session.send(r.request, proxies=env_proxies)

        # Add the previous requests to r.history so e.g. cookies get grabbed.
        r2.history.append(r)
        r2.history.append(login_r)

        return r2


class ElementSession(lemonpy.Session):

    headers = {
        "user-agent": "pycrunch/%s" % __version__
    }
    handler_class = ElementResponseHandler

    def __init__(self, email=None, password=None, token=None, domain=None,
                 progress_tracking=None):
        self.email = email
        self.password = password
        self.token = token
        self.domain = domain
        self.progress_tracking = progress_tracking or DefaultProgressTracking()
        super(ElementSession, self).__init__()



"""
User management
===============

Sharing a dataset
-----------------

To share a dataset with an existing user, you need the user's URL.
A common way to find that is to use the list of users in your account::

    account_users = site.user_url.account_url.users
    user = account_users.by('email')["bill.carson@mycorp.com"]
    print(user.dataset_permissions)
    pycrunch.elements.JSONObject(**{
        "edit": true,
        "change_permissions": true,
        "add_users": true,
        "change_weight": true,
        "view": true
    })

Note that each user has global limitations set by their account administrator;
you may not grant them permissions to any dataset that exceed these.

To grant or permissions you want that user to have on your dataset::

    myds.permissions.edit(
        user.entity_url,
        dataset_permissions={'view': True, 'edit': True}
    )

To grant to several users at once::

    myds.permissions.edit_index({
        user1.entity_url: {"dataset_permissions": {'view': True, 'edit': True}},
        user2.entity_url: {"dataset_permissions": {'view': True}},
        user3.entity_url: {"dataset_permissions": {'view': True}},
    })

The requesting user must have the 'add_users' and/or 'change_permissions'
permissions as necessary.

You can also set initial filters or a weight, and alter or revoke permissions.
See https://crunchio.atlassian.net/wiki/display/API/permissions for more info.
"""

import json


def invite(account, email, send_invite=None, url_base=None, id_method=None,
           role_url=None, account_permissions=None, **body):
    """Invite a new user directly to the given account. Return the new URL.

    id_method may be one of 'pwhash' (the default) or 'oauth'.
    role_url, if present, must be one of site.roles.accounts.
    """

    body['email'] = email
    if send_invite is not None:
        body['send_invite'] = send_invite
    if url_base is not None:
        body['url_base'] = url_base
    if id_method is not None:
        body['id_method'] = id_method
    if role_url is not None:
        body['role_url'] = role_url
    if account_permissions is not None:
        body['account_permissions'] = account_permissions

    return account.users.post(
        data=json.dumps({"element": "shoji:entity", "body": body})
    ).headers['Location']



#!/usr/bin/env python
# coding: utf-8

import os
import io
import re
from setuptools import setup, find_packages

thisdir = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(thisdir, 'src', 'pycrunch', 'version.py')) as v_file:
    VERSION = re.compile(
        r".*__version__ = '(.*?)'",
        re.S).match(v_file.read()).group(1)


def get_long_desc():
    readme_fn = os.path.join(thisdir, 'README.md')
    with io.open(readme_fn, encoding='utf-8') as stream:
        return stream.read()


requires = [
    'requests>=2.14.0',
    'six',
]

tests_requires = [
    'mock',
    'pytest',
    'pytest-cov',
]

setup_params = dict(
    name='pycrunch',
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
    version=VERSION,
    description="Crunch.io Client Library",
    long_description=get_long_desc(),
    url='https://github.com/Crunch-io/pycrunch',
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    author=u'Crunch.io',
    author_email='dev@crunch.io',
    license='LGPL',
    install_requires=requires,
    tests_require=tests_requires,
    packages=find_packages('src', exclude=['tests']),
    package_dir={'': 'src'},
    include_package_data=True,
    extras_require={
        'pandas:python_version=="3.4"': ['pandas~=0.19.0'],
        'pandas:python_version=="2.7" or python_version>="3.5"': ['pandas'],
        'testing:python_version=="3.4"': ['pandas~=0.19.0'],
        'testing:python_version=="2.7" or python_version>="3.5"': ['pandas'],
        'testing': tests_requires,
    },
    zip_safe=True,
    entry_points={},
)

if __name__ == '__main__':
    setup(**setup_params)



from unittest import TestCase

import requests
from pycrunch import Session, __version__
from pycrunch.lemonpy import ServerError

try:
    from requests.packages.urllib3.response import HTTPResponse
except ImportError:
    from urllib3.response import HTTPResponse
try:
    from unittest import mock
except ImportError:
    import mock


class TestHTTPRequests(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.s = Session("not an email", "not a password")
        cls.r = cls.s.get("http://httpbin.org/headers")

    def test_request_sends_user_agent(self):
        pycrunch_ua = 'pycrunch/%s' % __version__
        req_headers_sent = self.r.request.headers
        req_headers_received = self.r.json()['headers']
        self.assertTrue('user-agent' in req_headers_sent)
        self.assertTrue('User-Agent' in req_headers_received)
        self.assertTrue(pycrunch_ua in req_headers_sent.get('user-agent', ''))
        self.assertTrue(pycrunch_ua in req_headers_received.get('User-Agent', ''))

    def test_request_sends_gzip(self):
        req_headers_sent = self.r.request.headers
        req_headers_received = self.r.json()['headers']
        self.assertIn("gzip", req_headers_sent['Accept-Encoding'])
        self.assertIn("gzip", req_headers_received['Accept-Encoding'])


class TestHTTPResponses(TestCase):

    def test_response_with_no_content_type_header(self):
        # Simulate a 504 response with no Content-Type header and empty body.

        # pycrunch.lemonpy.ResponseHandler should be able to cope with this
        # peculiar situation.

        def _resp(adapter, request, *args, **kwargs):
            headers = {'Content-Length': '0'}
            response = HTTPResponse(status=504, body='', headers=headers)
            response = adapter.build_response(request, response)
            return response
        patch = mock.patch('requests.adapters.HTTPAdapter.send', _resp)

        patch.start()
        s = Session("not an email", "not a password")
        with self.assertRaises(ServerError) as exc_info:
            s.get("http://httpbin.org/status/504")
        patch.stop()

        response = exc_info.exception.args[0]
        assert isinstance(response, requests.models.Response)
        self.assertEqual(response.status_code, 504)

    def test_401_handle_calls_proxies(self):
        sess = Session("not an email", "not a password")
        headers = {'Set-Cookie': 'abx'}
        sess.post = lambda slf, *args, **kwargs: mock.MagicMock(headers=headers)
        sess.send = mock.MagicMock()
        url_401 = 'http://example.com/401'
        fake_request = mock.MagicMock(url=url_401)
        r = mock.MagicMock(
            request=fake_request,
            json=lambda: {'urls': {'login_url': 'http://www.httpbin.org/post'}}
        )

        from pycrunch.elements import ElementResponseHandler
        handler = ElementResponseHandler(sess)
        with mock.patch('pycrunch.elements.get_environ_proxies') as gep:
            gep.return_value = {}
            handler.status_401(r)
        gep.assert_called_once_with(url_401, no_proxy=None)
        sess.send.assert_called_with(fake_request, proxies={})



import json
import mimetypes
import os
import time

import six

from pycrunch import shoji, csvlib


class Importer(object):
    """A class for collecting the various ways to import data into Crunch.

    If the 'strict' argument is omitted or None, then any given CSV file
    is expected to conform to any previous metadata provided; that is,
    it shall not contain any columns nor categories not previously defined.
    If 0, then any undefined columns present in the CSV are ignored,
    and any undefined category ids generate new "missing" categories
    with the given id.
    """

    def __init__(self, retries=40, frequency=0.25,
                 backoff_rate=1.1, backoff_max=30, strict=None,
                 progress_tracker=None):
        self.retries = retries
        self.frequency = frequency
        self.backoff_rate = backoff_rate
        self.backoff_max = backoff_max
        self.strict = strict
        self.progress_tracker = progress_tracker

    def wait_for_batch_status(self, batch, status):
        """Wait for the given status(es) and return the batch. Error if not reached."""
        if isinstance(status, six.string_types):
            status = [status]

        for trial in range(self.retries):
            new_batch = batch.session.get(batch.self).payload
            st = new_batch.body['status']
            if st in ('error', 'failed'):
                raise ValueError("The batch was not fully appended.")
            elif st == 'conflict':
                raise ValueError("The batch had conflicts.")
            elif st in status:
                return new_batch
            else:
                time.sleep(self.frequency)
                if self.frequency < self.backoff_max:
                    self.frequency *= self.backoff_rate
                    if self.frequency > self.backoff_max:
                        self.frequency = self.backoff_max
        else:
            raise ValueError("The batch did not reach the '%s' state in the "
                             "given time. Please check again later." % status)

    def add_source(self, ds, filename, fp, mimetype):
        """Create a new Source on the given dataset and return its URL."""
        sources_url = ds.user_url.catalogs['sources']
        # Don't call Catalog.post here (which would force application/json);
        # we want requests.Session to set multipart/form-data with a boundary.
        new_source_url = ds.session.post(
            sources_url, files={"uploaded_file": (filename, fp, mimetype)}
        ).headers["Location"]

        if self.strict is not None:
            r = ds.session.get(new_source_url)
            if r.payload is None:
                raise TypeError("Response could not be parsed.", r)
            source = r.payload

            settings = source.body.get("settings", {})
            settings['strict'] = self.strict
            source.edit(settings=settings)

        return new_source_url

    def create_batch_from_source(self, ds, source_url, workflow=None,
                                 savepoint=True, autorollback=True):
        """Create and return a Batch on the given dataset for the given source."""
        batch = shoji.Entity(ds.session, body={
            'source': source_url,
            'workflow': workflow or []
        }, savepoint=savepoint, autorollback=autorollback)
        return ds.batches.create(batch, progress_tracker=self.progress_tracker).refresh()

    def append_rows(self, ds, rows):
        """Append the given rows of Python values. Return the new Batch."""
        f = csvlib.rows_as_csv_file(rows)
        return self.append_csv_string(ds, f)
    # Deprecated spelling:
    create_batch_from_rows = append_rows

    def append_csv_string(self, ds, csv_file, filename=None):
        """Append the given CSV string or open file. Return its Batch."""
        if filename is None:
            filename = 'upload.csv'

        source_url = self.add_source(ds, filename, csv_file, 'text/csv')
        return self.create_batch_from_source(ds, source_url)
    # Deprecated spellings:
    create_batch_from_csv_file = append_csv_string

    def append_stream(self, ds, fp, filename=None, mimetype=None):
        """Append the given file-like object to the dataset. Return its Batch."""
        if filename is None:
            filename = 'upload.crunch'

        if mimetype is None:
            mimetype, encoding = mimetypes.guess_type(filename)

        source_url = self.add_source(ds, filename, fp, mimetype)
        return self.create_batch_from_source(ds, source_url)

    def append_file(self, ds, path, filename=None, mimetype=None):
        """Append the file at the given path to the dataset. Return its Batch."""
        if filename is None:
            filename = path.rsplit(os.path.sep, 1)[-1]

        if mimetype is None:
            mimetype, encoding = mimetypes.guess_type(filename)

        source_url = self.add_source(ds, filename, open(path, 'rb'), mimetype)
        return self.create_batch_from_source(ds, source_url)

    def stream_rows(self, ds, values):
        """Send a data row (or list of rows) to the given dataset's stream.

        The rows are added to the Dataset's stream resource. This does
        *not* immediately append the rows to the Dataset; to do that,
        call append_pending_stream periodically.

        If the 'values' argument is a dict, it is treated as one row of
        {variable_id: data value} pairs. The given data values must be
        in the Crunch I/O format (for example, category ids instead
        of names or numeric_values). Otherwise, the 'values' argument
        must be an iterable of such dicts.
        """
        if isinstance(values, dict):
            values = [values]
        return ds.session.post(
            ds.fragments.stream,
            data="\n".join([json.dumps(row, indent=None) for row in values])
        )


importer = Importer()
"""A default Importer."""



"""
CSV support for Crunch.

The two functions, rows_as_csv_file* should be kept in sync.
The differing behavior cannot be factored out due to
the underlying stdlib csv module implementation and
performance characteristics of Python function calls.
"""

from __future__ import unicode_literals

import io
import csv
import six


def rows_as_csv_file(rows):
    """Return rows (iterable of lists of cells) as a CSV file open
    in binary mode.

    Any cells in the given
    rows which contain None will be emitted as an empty cell in the CSV
    (nothing between the commas), which Crunch interprets as {"?": -1},
    the "No Data" system missing value.
    """
    # Write to a StringIO because it joins lines as we go
    # and a .read() of it (like requests.post will do) does not make a copy.
    out = io.StringIO() if six.PY3 else io.BytesIO()

    sentinel = "__CSV_SENTINEL_NONE__"

    class EphemeralWriter():
        def write(self, line):
            line = line.replace(str('"' + sentinel + '"'), str())
            out.write(line)
    pipe = EphemeralWriter()

    writer = csv.writer(pipe, quoting=csv.QUOTE_NONNUMERIC, lineterminator='\n')
    for row in rows:
        row = [sentinel if cell is None else cell for cell in row]
        row = [
            cell.encode('utf-8')
            if six.PY2 and isinstance(cell, six.text_type) else cell
            for cell in row
        ]
        writer.writerow(row)

    out.seek(0)

    if six.PY3:
        out = io.BytesIO(out.getvalue().encode('utf-8'))

    return out


def rows_as_csv_file_clean(rows):
    """Return rows (iterable of lists of cells) as a CSV file open
    in binary mode.

    Duplicate of rows_as_csv_file except
    None values are emitted
    normally by the Python csv library, which means as (quoted) empty strings.
    Use this function if the rows contain no None values or
    if None values are only used for "text" types and the empty
    string is a suitable representation for those values.
    """
    # Write to a StringIO because it joins lines as we go
    # and a .read() of it (like requests.post will do) does not make a copy.
    out = io.StringIO() if six.PY3 else io.BytesIO()

    writer = csv.writer(out, quoting=csv.QUOTE_NONNUMERIC, lineterminator='\n')
    for row in rows:
        row = [
            cell.encode('utf-8')
            if six.PY2 and isinstance(cell, six.text_type) else cell
            for cell in row
        ]
        writer.writerow(row)

    out.seek(0)

    if six.PY3:
        out = io.BytesIO(out.getvalue().encode('utf-8'))

    return out



"""JSONObject and Element subclasses for Shoji objects.

See https://bitbucket.org/fumanchu/shoji/src/tip/spec.txt?at=default
for the latest Shoji specification.
"""

import json

import time
from six.moves import urllib

import six

from pycrunch import elements
from pycrunch.lemonpy import URL, ClientError, ServerError


class Tuple(elements.JSONObject):
    """A Shoji Tuple of attributes.

    Shoji Catalogs have an 'index' member, which maps URL's to Tuples.
    Shoji Entities have a 'body' member which is a Tuple.

    Like all JSONObjects, the items in a Tuple are readable as keys
    (e.g. tup['foo']) or as attributes (e.g. tup.foo). In addition,
    the URL of the Entity (whether from Catalog.index or Entity.self)
    is included as Tuple.entity_url. The Tuple.fetch method then assumes
    .entity_url as the URL to request, and either returns the complete
    Entity or raises TypeError if the response could not be parsed.
    """

    def __init__(self, session, entity_url, **members):
        self.session = session
        self.entity_url = entity_url
        self._entity = None
        elements.JSONObject.__init__(self, **members)

    def copy(self):
        """Return a (shallow) copy of self."""
        return self.__class__(self.session, self.entity_url, **self)

    def fetch(self, *args, **kwargs):
        r = self.session.get(self.entity_url.absolute, *args, **kwargs)
        if r.payload is None:
            raise TypeError("Response could not be parsed.", r)
        return r.payload

    @property
    def entity(self):
        """Fetch, cache, and return the shoji.Entity for self.entity_url.

        This is typically used from a Catalog Tuple to GET the associated
        Entity body attributes; e.g. foo_catalog['bar'].entity.body.qux.
        However, it can also be used from an Entity.body tuple to obtain
        a copy of the whole Entity; e.g. bar2 = bar.body.entity.
        """
        if self._entity is None:
            self._entity = self.fetch()
        return self._entity


class Index(elements.JSONObject):
    """The index of a Shoji Catalog.

    Shoji Catalogs have an 'index' member, which maps lemonpy.URL's to Tuples.
    """

    def __init__(self, session, catalog_url, **members):
        self.session = session
        self.catalog_url = catalog_url
        catalog_url_absolute, frag = urllib.parse.urldefrag(catalog_url.absolute)

        # Normalized keys stores a dictionary of shortest relative URLs
        # obtained via URL.relative_to that maps to the arbitrary URL relative
        # or absolute that the client submitted. We can use this to always
        # key the tuples by relative URL regardless of what kind of URL the
        # user asks to fetch an item by.
        normalized_keys = {}
        for entity_url, tup in six.iteritems(members):
            if tup is not None:
                url = entity_url
                if not hasattr(url, "relative_to"):  # Faster than isinstance(url, URL)
                    url = URL(url, catalog_url_absolute)

                members[url] = Tuple(session, url, **tup)
                normalized_keys[url.relative_to(catalog_url)] = url

        self.normalized_keys = normalized_keys
        elements.JSONObject.__init__(self, **members)

    def _rel_abs(self, url):
        if not hasattr(url, 'relative_to'):
            cat_url = self.catalog_url
            if '://' in url:
                abs_url = URL(url, cat_url)
            else:
                abs_url = URL(URL(url, cat_url).absolute, cat_url)
            rel_url = abs_url.relative_to(cat_url)
        else:
            abs_url = url.absolute
            rel_url = url.relative_to(self.catalog_url)
        return rel_url, abs_url

    def __getitem__(self, item):
        """
        We will always convert the received key to a relative URL using
        URL.relative_to so we can use self.normalized_keys to know which
        URL the user sent to key this tuple by.
        """
        rel_url, abs_url = self._rel_abs(item)
        key = self.normalized_keys[rel_url]
        return super(Index, self).__getitem__(key)

    def __setitem__(self, key, value):
        rel_url, abs_url = self._rel_abs(key)
        if isinstance(value, dict):
            value = Tuple(self.session, abs_url, **value)
        self.normalized_keys[rel_url] = key
        return super(Index, self).__setitem__(key, value)

    def update(self, mapping):
        for k, v in mapping.items():
            self[k] = v  # To go through __setitem__'s logic


class CreateMixin(object):

    def create(self, entity=None, progress_tracker=None):
        """POST the given Entity to this catalog to create a new resource.

        The 'entity' arg may be a complete shoji.Entity, in which case
        its .json will be POST'ed to this Catalog, or it may be a plain
        dict of attributes, in which case it will first be wrapped in an
        Entity. This latter case is more common simply for the fact that
        it results in cleaner calling code; compare:

            foo_catalog.create(pycrunch.shoji.Entity(my.session, bar=qux))

        versus:

            foo_catalog.create({"bar": qux})

        The `progress_tracker` argument allows overriding progress tracking
        configuration provided by session.
        See ``Entity.wait_progress`` for details.

        An entity is returned.
        """
        if entity is None:
            entity = Entity(self.session)
        elif isinstance(entity, dict) and not isinstance(entity, elements.Document):
            entity = Entity(self.session, **entity)

        response = self.post(data=entity.json)
        if response.history:
            # Coming from a redirect
            seeother = [r for r in response.history if r.status_code == 303]
            if seeother:
                # We got a See Other response, this means the resource
                # was not created because an equivalent one is already available.
                entity['self'] = URL(seeother[-1].headers['Location'], '')
                return entity
        return self._wait_for_progress(entity, response, progress_tracker)

    def _wait_for_progress(self, entity, r, progress_tracker):
        entity['self'] = URL(r.headers['Location'], '')
        if r.status_code == 202:
            try:
                progress_url = r.payload['value']  # noqa
            except Exception:
                # Not a progress API just return the incomplete entity.
                # User will refresh it.
                pass
            else:
                # We have a progress_url, wait for completion
                entity.wait_progress(r, progress_tracker)
        return entity

    def by(self, attr):
        """Return the Tuples of self.index indexed by the given 'attr' instead.

        If a given Tuple does not contain the specified attribute,
        it is not included. If more than one does, only one will be
        included (which one is undefined).

        The specified attr is not popped from the Tuple; it is merely
        copied to the output keys. Due to restrictions on Python dicts,
        specifying attrs which are not hashable will raise an error.
        """
        return elements.JSONObject(
            (tupl[attr], tupl)
            for tupl in six.itervalues(self.index)
            if attr in tupl
        )


class Catalog(elements.Document, CreateMixin):
    """A Shoji Catalog."""

    element = "shoji:catalog"
    navigation_collections = ("catalogs", "orders", "views", "urls")

    def __init__(__this__, session, **members):
        if 'self' in members:
            if not isinstance(members['self'], URL):
                members['self'] = URL(members['self'], "")
            if 'index' in members:
                members['index'] = Index(session, members['self'], **members['index'])
        super(Catalog, __this__).__init__(session, **members)

    def add(self, entity_url, attrs=None, **kwargs):
        """Add the given entity, plus any spurious index attributes (ICK), to self.

        This is a total hack because Crunch has an endpoint (dataset permissions)
        where non-tuples are included in a Catalog.index.
        """
        kwargs[entity_url] = attrs or {}
        p = json.dumps(dict(element="shoji:catalog", self=self.self, index=kwargs))
        return self.patch(data=p).payload

    def edit(self, entity_url, **attrs):
        """Update the catalog with the given entity attributes."""
        p = self.__class__(self.session, self=self.self, index={entity_url: attrs})
        return self.patch(data=p.json).payload

    def edit_index(self, index):
        """Update the catalog with the given (probably partial) index."""
        p = self.__class__(self.session, self=self.self, index=index)
        return self.patch(data=p.json).payload

    def drop(self, entity_url):
        """Delete the given entity from the catalog."""
        p = self.__class__(self.session, self=self.self, index={entity_url: None})
        return self.patch(data=p.json).payload


class Entity(elements.Document, CreateMixin):

    element = "shoji:entity"
    navigation_collections = ("catalogs", "fragments", "views", "urls", 'orders')

    def __init__(__this__, session, **members):
        members.setdefault("body", {})
        if 'self' in members:
            if not isinstance(members['self'], URL):
                members['self'] = URL(members['self'], "")
            members['body'] = Tuple(session, members['self'], **members['body'])
            if 'index' in members:
                members['index'] = Index(session, members['self'], **members['index'])
        super(Entity, __this__).__init__(session, **members)

    def edit(self, **body_attrs):
        """Update the entity with the new body attributes."""
        p = self.__class__(self.session, body=body_attrs)
        payload = super(Entity, self).patch(data=p.json).payload
        self.body.update(body_attrs)
        return payload

    def replace(self, entity=None):
        """PUT the given entity (default: self) to self.url."""
        if entity is None:
            entity = self
        return super(Entity, self).put(data=entity.json).payload

    def wait_progress(self, r, progress_tracker=None):
        self['self'] = URL(r.headers['Location'], '')
        wait_progress(r, self.session, progress_tracker, entity=self)
        return self


def wait_progress(r, session, progress_tracker=None, entity=None):
    """Waits for completion of an Entity or View from API response
    that provides progress reporting.

    The entity will be updated with the location provided by the response
    and the method will wait until progress completed or progress tracker
    did timeout.

    User need to manually call ``.refresh()`` when the progress completed
    to fetch the updated entity.

    A custom ``progress_tracker`` can be passed to override the one
    provided by session. Progress trackers provide a way to configure
    timeout, polling interval and reporting callbacks.

    See `pycrunch.progress.DefaultProgressTracking` and
    `pycrunch.progress.SimpleTextBarProgressTracking` for documentation
    regarding progress trackers.
    """
    progress_url = r.payload['value']

    if progress_tracker is None:
        progress_tracker = session.progress_tracking

    timeout = progress_tracker.timeout
    progress_state = progress_tracker.start_progress()
    begin = time.time()
    while timeout is None or time.time() - begin < timeout:
        prog_r = session.get(progress_url)
        progress = prog_r.payload['value']
        progress_tracker.on_progress(progress_state, progress)
        if progress['progress'] == -1:
            # Completed due to error
            raise TaskError(progress['message'])
        elif progress['progress'] == 100:
            # Completed with success
            break
        time.sleep(progress_tracker.interval)
    else:
        # Loop completed due to timeout
        raise TaskProgressTimeoutError(entity, r, timeout=timeout)


class View(elements.Document):

    element = "shoji:view"
    navigation_collections = ("views", "urls")

    def __init__(__this__, session, **members):
        if 'self' in members and not isinstance(members['self'], URL):
            members['self'] = URL(members['self'], "")
        super(View, __this__).__init__(session, **members)

    @property
    def value(self):
        return self['value']

    @value.setter
    def value(self, newvalue):
        """Update the View with the new value."""
        self['value'] = newvalue
        super(View, self).put(data=self.json)

    def wait_progress(self, r, progress_tracker=None):
        self['self'] = URL(r.headers['Location'], '')
        wait_progress(r, self.session, progress_tracker, entity=self)
        return self


class Order(elements.Document):

    element = "shoji:order"
    navigation_collections = ("catalogs",)

    def __init__(__this__, session, **members):
        if 'self' in members and not isinstance(members['self'], URL):
            members['self'] = URL(members['self'], "")
        super(Order, __this__).__init__(session, **members)

    @property
    def graph(self):
        return self['graph']

    @graph.setter
    def graph(self, newgraph):
        """Update the Order with the new graph."""
        self['graph'] = newgraph
        super(Order, self).put(data=self.json)


class TaskProgressTimeoutError(Exception):
    def __init__(self, entity, response, timeout=''):
        if timeout:
            timeout = '%s seconds ' % (timeout, )
        super(TaskProgressTimeoutError, self).__init__(
            'Task Progress did not complete before {}timeout. '
            'Trap this exception and call exc.entity.wait_progress(exc.response) '
            'to wait for completion explicitly.'.format(timeout)
        )
        self.entity = entity
        self.response = response


class TaskError(ClientError, ServerError):
    # For backward compatibility inherit from ClientError and ServerError

    def __init__(self, message):
        Exception.__init__(self, message)

    @property
    def status_code(self):
        return None

    @property
    def message(self):
        return self.args[0]
